from . import utils
from typing import *
from . import buttons


class BotCommand:
    def __init__(self, name: str, description: str):
        self.name = name
        self.description = description

    
    def as_dict(self):
        return {
            "name": self.name, "description": self.description
        }


class User:
    def __init__(self,
        user_id: int,
        first_name: str,
        name: str,
        is_bot: bool,
        last_activity_time: int,
        last_name: "str | None" = None,
        username: "str | None" = None,
        description: "str | None" = None,
        avatar_url: "str | None" = None,
        full_avatar_url: "str | None" = None,
        commands: "List[BotCommand] | None" = None,
        last_access_time: "int | None" = None,
        is_owner: "bool | None" = None,
        is_admin: "bool | None" = None,
        join_time: "int | None" = None,
        permissions: "List[str] | None" = None

    ):
        self.user_id: int = user_id
        self.first_name: str = first_name
        self.last_name: str = last_name
        self.name: str = name
        self.username: "str | None" = username
        self.is_bot: bool = is_bot
        self.last_activity_time: int = last_activity_time
        self.description: "str | None" = description
        self.avatar_url: "str | None" = avatar_url
        self.full_avatar_url: "str | None" = full_avatar_url
        self.commands: "List[BotCommand] | None" = [
            BotCommand(**i) for i in commands
        ] if commands else None
        self.last_access_time: "int | None" = last_access_time
        self.is_owner: "bool | None" = is_owner
        self.is_admin: "bool | None" = is_admin
        self.join_time: "int | None" = join_time
        self.permissions: "List[str] | None" = permissions


    def __repr__(self):
        return f"{type(self).__name__}(user_id={self.user_id!r}, name={self.name!r})"


    def __eq__(self, other):
        if isinstance(other, User):
            return self.user_id == other.user_id
        return False


    @staticmethod
    def from_json(data: dict) -> "User | None":
        if data == None: return None

        return User(**data)


class Attachment:
    def __init__(self, type: str):
        self.type: str = type


    @staticmethod
    def from_json(data: dict) -> "Attachment | None":
        if data['type'] == 'image':
            return PhotoAttachment.from_json(data)
        elif data['type'] == 'video':
            return VideoAttachment.from_json(data)
        elif data['type'] == 'audio':
            return AudioAttachment.from_json(data)
        elif data['type'] == 'file':
            return FileAttachment.from_json(data)
        elif data['type'] == 'sticker':
            return StickerAttachment.from_json(data)
        elif data['type'] == 'contact':
            return ContactAttachment.from_json(data)
        elif data['type'] == 'share':
            return ShareAttachment.from_json(data)
        elif data['type'] == 'location':
            return LocationAttachment.from_json(data)
        elif data['type'] == 'inline_keyboard':
            return InlineKeyboardAttachment.from_json(data)
        else:
            raise Exception(f"Unknown attachment type: {data['type']}")


class MediaPayload:
    def __init__(self,
        token: str,
        url: "str | None" = None,
    ):
        self.url: "str | None" = url
        self.token: str = token


    @staticmethod
    def from_json(data: dict) -> "MediaPayload | None":
        return MediaPayload(url=data.get('url'), token=data.get('token'))


class StickerPayload:
    def __init__(self,
        url: str,
        code: str,
    ):
        self.url: str = url
        self.code: str = code


    @staticmethod
    def from_json(data: dict) -> "StickerPayload | None":
        return StickerPayload(data['url'], data['code'])


class ContactPayload:
    def __init__(self,
        vcf_info: "str | None" = None,
        max_info: "User | None" = None,
    ):
        self.vcf_info: "str | None" = vcf_info
        self.max_info: "User | None" = max_info


    @staticmethod
    def from_json(data: dict) -> "ContactPayload | None":
        return ContactPayload(data['vcf_info'], data['max_info'])


class PhotoPayload(MediaPayload):
    def __init__(self,
        token: str,
        url: "str | None" = None,
        photo_id: "int | None" = None
    ):
        super().__init__(url=url, token=token)
        self.photo_id: "int | None" = photo_id

    
    @staticmethod
    def from_json(data: dict) -> "PhotoPayload | None":
        return PhotoPayload(url=data.get('url'), token=data.get('token'), photo_id=data.get('photo_id'))


class PhotoAttachment(Attachment):
    def __init__(self,
        payload: PhotoPayload
    ):
        super().__init__("image")
        self.payload: PhotoPayload = payload


    @staticmethod
    def from_json(data: dict) -> "PhotoAttachment | None":
        return PhotoAttachment(
            PhotoPayload.from_json(data['payload'])
        )
    
    
    def as_dict(self):
        return {
            'type': self.type,
            'payload': {'token': self.payload.token}
        }


class VideoAttachment(Attachment):
    def __init__(self,
        payload: MediaPayload,
        thumbnail: "str | None" = None,
        width: "int | None" = None,
        height: "int | None" = None,
        duration: "int | None" = None
    ):
        super().__init__("video")
        self.payload: MediaPayload = payload
        self.thumbnail: "str | None" = thumbnail
        self.width: "int | None" = width
        self.height: "int | None" = height
        self.duration: "int | None" = duration

    
    @staticmethod
    def from_json(data: dict) -> "VideoAttachment | None":
        return VideoAttachment(
            MediaPayload.from_json(data['payload']),
            data.get('thumbnail', None),
            data.get('width', None),
            data.get('height', None),
            data.get('duration', None),
        )
    
    
    def as_dict(self):
        return {
            'type': self.type,
            'payload': {'token': self.payload.token}
        }


class AudioAttachment(Attachment):
    def __init__(self,
        payload: MediaPayload,
        transcription: "str | None" = None
    ):
        super().__init__("audio")
        self.payload: MediaPayload = payload
        self.transcription: "str | None" = transcription


    @staticmethod
    def from_json(data: dict) -> "AudioAttachment | None":
        return AudioAttachment(
            MediaPayload.from_json(data['payload']),
            data.get('transcription', None)
        )
    
    
    def as_dict(self):
        return {
            'type': self.type,
            'payload': {'token': self.payload.token}
        }


class FileAttachment(Attachment):
    def __init__(self,
        payload: MediaPayload,
        filename: "str | None" = None,
        size: "int | None" = None
    ):
        super().__init__("file")
        self.payload: MediaPayload = payload
        self.filename: "str | None" = filename
        self.size: "int | None" = size


    @staticmethod
    def from_json(data: dict) -> "FileAttachment | None":
        return FileAttachment(
            MediaPayload.from_json(data['payload']),
            data.get('filename'),
            data.get('size')
        )
    
    
    def as_dict(self):
        return {
            'type': self.type,
            'payload': {'token': self.payload.token}
        }


class StickerAttachment(Attachment):
    def __init__(self,
        payload: StickerPayload,
        width: "int | None" = None,
        height: "int | None" = None
    ):
        super().__init__("sticker")
        self.payload: StickerPayload = payload
        self.width: int = width
        self.height: int = height

    
    @staticmethod
    def from_json(data: dict) -> "StickerAttachment | None":
        return StickerAttachment(
            StickerPayload.from_json(data['payload']),
            data.get('width', None),
            data.get('height', None)
        )


class ContactAttachment(Attachment):
    def __init__(self,
        payload: ContactPayload,
    ):
        super().__init__("contact")
        self.payload: ContactPayload = payload


    @staticmethod
    def from_json(data: dict) -> "ContactAttachment | None":
        return ContactAttachment(
            ContactPayload.from_json(data['payload'])
        )


class ShareAttachment(Attachment):
    def __init__(self,
        payload: "MediaPayload | None",
        url: "str | None" = None,
        title: "str | None" = None,
        description: "str | None" = None,
        image_url: "str | None" = None,
    ):
        super().__init__("share")
        self.url: "str | None" = url
        self.payload: "MediaPayload | None" = payload
        self.title: "str | None" = title
        self.description: "str | None" = description
        self.image_url: "str | None" = image_url


    @staticmethod
    def from_json(data: dict) -> "ShareAttachment | None":
        return ShareAttachment(
            MediaPayload.from_json(data.get('payload', None)),
            data['payload'].get('url', None),
            data.get('title', None),
            data.get('description', None),
            data.get('image_url', None),
        )


class LocationAttachment(Attachment):
    def __init__(self,
        latitude: float,
        longitude: float,
    ):
        super().__init__("location")
        self.latitude: float = latitude
        self.longitude: float = longitude


    @staticmethod
    def from_json(data: dict) -> "LocationAttachment | None":
        return LocationAttachment(
            data['latitude'],
            data['longitude']
        )


class InlineKeyboardAttachment(Attachment):
    def __init__(self,
        payload: List[List[buttons.Button]],
    ):
        super().__init__("inline_keyboard")
        self.payload: List[List[buttons.Button]] = payload


    @staticmethod
    def from_json(data: dict) -> "InlineKeyboardAttachment | None":
        return InlineKeyboardAttachment(
            [[buttons.Button.from_json(j) for j in i] for i in data['payload']['buttons']]
        )


class MessageRecipient:
    def __init__(self,
        chat_id: "int | None",
        chat_type: str
    ):
        self.chat_id: "int | None" = chat_id
        self.chat_type: str = chat_type

    def __repr__(self):
        return f"{type(self).__name__}(chat_id={self.chat_id!r}, chat_type={self.chat_type!r})"
    
    def __eq__(self, other):
        if isinstance(other, MessageRecipient):
            return self.chat_id == other.chat_id
        return False

    @staticmethod
    def from_json(data: dict) -> "MessageRecipient":
        if data == None: return None

        return MessageRecipient(
            chat_id = data["chat_id"],
            chat_type = data["chat_type"]
        )
    

class Markup:
    def __init__(self,
        type: Literal[
            'strong', 'emphasized', 'monospaced', 'link', 'strikethrough',
            'underline', 'user_mention', 'heading', 'highlighted'
        ],
        start: int,
        length: int,
        user_link: "str | None" = None,
        user_id: "int | None" = None,
        url: "str | None" = None
    ):
        '''
        A markup element

        :param type: Markup type
        :param start: Start position
        :param length: Length
        :param user_link: Username. `None` if markup type is not `user_link`
        :param user_id: User ID. `None` if markup type is not `user_link`
        :param url: URL. `None` if markup type is not `link`
        '''
        self.type: Literal[
            'strong', 'emphasized', 'monospaced', 'link', 'strikethrough',
            'underline', 'user_mention', 'heading', 'highlighted'
        ] = type
        self.start: int = start
        self.length: int = length

        self.user_link: "str | None" = user_link
        self.user_id: "int | None" = user_id
        self.url: "str | None" = url


    @staticmethod
    def from_json(data: dict) -> "Markup | None":
        if data == None: return None

        if data['type'] == 'user_mention':
            return Markup(
                data['type'], data['from'], data['length'],
                user_link=data.get('user_link', None), user_id=data.get('user_id', None)
            )
        elif data['type'] == 'link':
            return Markup(
                data['type'], data['from'], data['length'], url=data['url']
            )

        return Markup(data['type'], data['from'], data['length'])


class MessageBody:
    def __init__(self,
        mid: str,
        seq: int,
        text: "str | None",
        attachments: "List[Attachment] | None",
        markup: "List[Markup] | None" = None
    ):
        self.message_id: str = mid
        self.seq: int = seq
        self.text: "str | None" = text
        self.attachments: "List[Attachment] | None" = attachments
        self.markup: "List[Markup] | None" = markup


    @staticmethod
    def from_json(data: dict) -> "MessageBody":
        if data == None: return None

        return MessageBody(
            mid = data["mid"],
            seq = data["seq"],
            text = data["text"],
            attachments = [Attachment.from_json(x) for x in data.get('attachments', [])],
            markup = [Markup.from_json(x) for x in data.get('markup', [])]
        )


class LinkedMessage:
    def __init__(self,
        type: str,
        message: MessageBody,
        sender: User,
        chat_id: "int | None" = None,
    ):
        self.type: str = type
        self.message: MessageBody = message
        self.sender: User = sender
        self.chat_id: "int | None" = chat_id


    @staticmethod
    def from_json(data: dict) -> "LinkedMessage":
        if data == None: return None
            
        return LinkedMessage(
            type = data["type"],
            message = MessageBody.from_json(data["message"]),
            sender = User.from_json(data.get('sender', None)),
            chat_id = data.get('chat_id', None),
        )
    
    @property
    def user_id(self):
        return self.sender.user_id


class Message:
    def __init__(self,
        recipient: MessageRecipient,
        body: MessageBody,
        timestamp: float,
        sender: User,
        link: "LinkedMessage | None" = None,
        views: "int | None" = None,
        url: "str | None" = None,
        bot = None
    ):
        self.recipient: MessageRecipient = recipient
        self.body: "MessageBody | None" = body
        self.timestamp: "float | None" = timestamp
        self.sender: "User | None" = sender
        self.link: "LinkedMessage | None" = link
        self.views: "int | None" = views
        self.url: "str | None" = url
        self.user_locale: "str | None" = None
        self.bot = bot

    def __repr__(self):
        return f"{type(self).__name__}(text={self.body.text!r})"

    def __str__(self):
        return self.body.text
    
    def __eq__(self, other):
        if isinstance(other, Message):
            return self.id == other.id
        return False
    
    @property
    def id(self) -> str:
        return self.body.message_id
    
    @property
    def content(self) -> str:
        return self.body.text
    
    @property
    def user_id(self):
        return self.sender.user_id

    @staticmethod
    def from_json(data: dict) -> "Message":
        return Message(
            recipient = MessageRecipient.from_json(data.get('recipient')),
            body = MessageBody.from_json(data.get('body', None)),
            timestamp = data.get('timestamp', None),
            sender = User.from_json(data.get("sender", None)),
            link = LinkedMessage.from_json(data.get("link", None)),
            views = data.get("stat", {}).get("views", None),
            url = data.get("url", None)
        )
    

    def resolve_mention(self, replies: bool = True, message_text: bool = True, skip_bot: bool = True) -> "int | None":
        '''
        Finds who was mentioned in this message and returns the user ID if found.

        :param replies: Whether to check for this message's link author.
        :param message_text: Whether to check for mentions in message text.
        :param skip_bot: Whether to ignore mentions of the bot.
        '''
        if replies and self.link:
            if self.link.type == 'reply':
                if skip_bot and self.bot and self.link.sender.user_id == self.bot.id:
                    pass
                else:
                    return self.link.sender.user_id
            
        if message_text:
            for i in self.body.markup:
                if i.type != 'user_mention': continue
                if skip_bot and self.bot and i.user_id == self.bot.id: continue
                if skip_bot and self.bot and i.user_link == self.bot.username: continue
                return i.user_id


    async def send(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> "Message":
        '''
        Send a message to the chat that the message is sent.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        if self.bot == None:
            return
        return (await self.bot.send_message(
            text, chat_id=self.recipient.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments
        ))


    async def reply(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> "Message":
        '''
        Reply to this message.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        if self.bot == None:
            return
        return (await self.bot.send_message(
            text, chat_id=self.recipient.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments, reply_to=self.id
        ))
    

    async def edit(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        reply_to: "int | None" = None,
        notify: bool = True,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
    ) -> "Message":
        '''
        Edit a message

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        if self.bot == None:
            return
        return (await self.bot.edit_message(
            self.id, text,
            format=format, notify=notify,
            keyboard=keyboard, reply_to=reply_to
        ))
    
    async def delete(self):
        if self.bot == None:
            return
        return await self.bot.delete_message(
            self.id
        )


class BotStartPayload:
    def __init__(self,
        chat_id: int,
        user: User,
        payload: "str | None",
        user_locale: "str | None",
        bot = None
    ):
        self.chat_id: int = chat_id
        self.user: User = user
        self.payload: "str | None" = payload
        self.user_locale: "str | None" = user_locale
        self.bot = bot


    @staticmethod
    def from_json(data: dict, bot) -> "BotStartPayload":
        return BotStartPayload(
            chat_id = data["chat_id"],
            user = User.from_json(data["user"]),
            payload = data.get('payload', None),
            user_locale = data.get('user_locale', None),
            bot = bot,
        )
    
    async def send(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> "Message":
        '''
        Send a message to the chat where bot was started.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        if self.bot == None:
            return
        return (await self.bot.send_message(
            text, chat_id=self.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments
        ))
    
    @property
    def user_id(self):
        return self.user.user_id
    
    

class CommandContext:
    def __init__(self,
        bot,
        message: Message,
        command_name: str,
        args: str
    ):
        self.bot = bot
        self.message: Message = message
        self.sender: User = message.sender
        self.recipient: MessageRecipient = message.recipient
        self.command_name: str = command_name
        self.args_raw: str = args
        self.args: List[str] = args.split()


    async def send(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> Message:
        '''
        Send a message to the chat that the user sent the command.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        return (await self.bot.send_message(
            text, chat_id=self.message.recipient.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments
        ))


    async def reply(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> Message:
        '''
        Reply to the message that the user sent.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        return (await self.bot.send_message(
            text, chat_id=self.message.recipient.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments, reply_to=self.message.id
        ))
    
    @property
    def user_id(self):
        return self.sender.user_id


class CommandHandler:
    def __init__(
        self,
        call: Callable,
        as_message: bool = False,
    ):
        self.call = call
        self.as_message: bool = as_message


class Handler:
    def __init__(
        self,
        call: Callable,
        deco_filter: "Callable | None" = None,
        router_filters: List[Callable] = [],
    ):
        self.call = call
        self.deco_filter: "Callable | None" = deco_filter
        self.router_filters: List[Callable] = router_filters

    
    @property
    def filters(self) -> List[Callable]:
        if self.deco_filter:
            return [self.deco_filter, *self.router_filters]
        return self.router_filters


class Image:
    def __init__(self,
        url: str,
    ):
        '''
        An image.

        :param url: Image URL
        '''
        self.url: str = url
    
    
    @staticmethod
    def from_json(data: dict) -> "Image | None":
        if data == None: return None

        return Image(**data)


class ImageRequestPayload:
    def __init__(self,
        url: "str | None" = None,
        token: "str | None" = None
    ):
        '''
        A payload with the info about an image or avatar to send to the bot.

        Only url or token must be specified.
        
        :param url: Image URL
        :param token: Attachment token generated by Bot.upload_image().token
        '''
        assert url or token, 'Token or URL must be specified'
        assert not (url and token), 'Token and URL cannot be specified at the same time'
        
        self.url: "str | None" = url
        self.token: "str | None" = token
    
    
    @staticmethod
    def from_json(data: dict) -> "ImageRequestPayload | None":
        if data == None: return None

        return ImageRequestPayload(**data)
    

    def as_dict(self):
        return { "url": self.url } if self.url else { "token": self.token }
    

class Chat:
    def __init__(self,
        chat_id: int,
        type: str,
        status: str,
        last_event_time: int,
        participants_count: int,
        is_public: bool,
        title: "str | None" = None,
        icon: "Image | None" = None,
        description: "str | None" = None,
        pinned_message: "Message | None" = None,
        owner_id: "int | None" = None,
        participants: "Dict[str, int] | None" = None,
        link: "str | None" = None,
        messages_count: "str | None" = None,
        chat_message_id: "str | None" = None,
        dialog_with_user: "User | None" = None,
    ):
        self.chat_id: int = chat_id
        self.type: str = type
        self.status: str = status
        self.last_event_time: int = last_event_time
        self.participants_count: int = participants_count
        self.title: "str | None" = title
        self.icon: "Image | None" = icon
        self.is_public: bool = is_public
        self.dialog_with_user: "User | None" = dialog_with_user
        self.description: "str | None" = description
        self.pinned_message: "Message | None" = pinned_message
        self.owner_id: "int | None" = owner_id
        self.participants: "Dict[int, int] | None" = {int(k): v for k, v in participants.items()} if participants else None
        self.link: "str | None" = link
        self.messages_count: "str | None" = messages_count
        self.chat_message_id: "str | None" = chat_message_id

    def __eq__(self, other):
        if isinstance(other, Chat):
            return self.chat_id == other.chat_id
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}(chat_id={self.chat_id!r}, title={self.title!r})"

    @staticmethod
    def from_json(data: dict) -> "Chat | None":
        if data == None: return None

        return Chat(**data)
    

class Callback:
    def __init__(self,
        bot,
        timestamp: int,
        callback_id: str,
        message: "Message | None",
        user: User,
        user_locale: "str | None",
        payload: "str | None" = None
    ):
        self.bot = bot
        self.timestamp: int = timestamp
        self.callback_id: str = callback_id
        self.message: "Message | None" = message
        self.user: User = user
        self.payload: "str | None" = payload
        self.user_locale: "str | None" = user_locale

    
    @property
    def content(self) -> str:
        return self.payload


    async def send(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> "Message":
        '''
        Send a message to the chat that contains the message with the pressed button.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        if self.bot == None:
            return
        assert self.message != None, 'Original message not found'
        
        return (await self.bot.send_message(
            text, chat_id=self.message.recipient.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments
        ))


    async def reply(self,
        text: str,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> "Message":
        '''
        Reply to the message with the button.

        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link preview. False by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        if self.bot == None:
            return
        assert self.message != None, 'Original message not found'
        
        return (await self.bot.send_message(
            text, chat_id=self.message.recipient.chat_id,
            format=format, notify=notify, disable_link_preview=disable_link_preview,
            keyboard=keyboard, attachments=attachments, reply_to=self.message.id
        ))
    

    async def answer(self,
        notification: "str | None" = None,
        text: "str | None" = None,
        format: "Literal['html', 'markdown', 'default'] | None" = 'default',
        notify: bool = True,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ):
        '''
        Answer the callback.

        :param notification: Notification to display to the user
        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param notify: Whether to notify users about the message. True by default.
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        assert notification != None or text != None or attachments,\
            'Either notification, text or attachments must be specified'
        body = {
            'notification': notification,
            'message': None
        }
        if keyboard == None and self.message != None:
            keyboard = [i for i in self.message.body.attachments if i.type == 'inline_keyboard']
            if len(keyboard) == 0:
                keyboard = None
            else:
                keyboard = keyboard[0].payload

        if text != None:
            format = self.bot.default_format if format == 'default' else format
            body['message'] = utils.get_message_body(text, format, notify=notify, keyboard=keyboard, attachments=attachments)

        out = await self.bot.post(
            'https://botapi.max.ru/answers', params={'callback_id': self.callback_id},
            json=body
        )
        return await out.json()
    

    @property
    def user_id(self):
        return self.user.user_id


    @staticmethod
    def from_json(data: dict, message: "dict | None", user_locale: "str | None" = None, bot = None) -> "Callback | None":
        if data == None: return None
        
        return Callback(
            bot,
            data['timestamp'],
            data['callback_id'],
            Message.from_json(message) if message != None else None,
            User.from_json(data['user']),
            user_locale,
            data.get('payload', None)
        )


class ChatCreatePayload:
    def __init__(self,
        timestamp: int,
        chat: Chat,
        message_id: "str | None" = None,
        start_payload: "str | None" = None
    ):
        '''
        Payload that is sent to the `Bot.on_button_chat_create` decorator.

        :param timestamp: Timestamp of the button press
        :param chat: Created chat
        :param message_id: Message ID on which the button was
        :param start_payload: Start payload specified by the button
        '''
        self.timestamp: int = timestamp
        self.chat: Chat = chat
        self.message_id: "str | None" = message_id
        self.start_payload: "str | None" = start_payload


    @staticmethod
    def from_json(data: dict) -> "ChatCreatePayload | None":
        if data == None: return None
        
        return ChatCreatePayload(
            data['timestamp'],
            Chat.from_json(data['chat']),
            data.get('message_id', None),
            data.get('start_payload', None)
        )
    


class MessageDeletePayload:
    def __init__(self,
        timestamp: int,
        message: "Message | None" = None,
        message_id: "str | None" = None,
        chat_id: "int | None" = None,
        user_id: "int | None" = None,
        bot = None
    ):
        '''
        Payload that is sent to the `Bot.on_message_delete` decorator.

        :param timestamp: Timestamp of the message deletion.
        :param message: Cached Message object. May be None if message was not cached
        :param message_id: ID of the deleted message
        :param chat_id: ID of the chat the message was deleted in
        :param user_id: ID of the user who deleted the message
        '''
        self.timestamp: int = timestamp
        self.message: "Message | None" = message
        self.message_id: "str | None" = message_id
        self.chat_id: "int | None" = chat_id
        self.user_id: "int | None" = user_id
        self.bot = bot


    @staticmethod
    def from_json(data: dict, bot) -> "MessageDeletePayload | None":
        if data == None: return None
        
        return MessageDeletePayload(
            data['timestamp'],
            bot.cache.get_message(data.get('message_id', None)),
            data.get('message_id', None),
            data.get('chat_id', None),
            data.get('user_id', None),
            bot = bot
        )


    @property
    def content(self) -> "str | None":
        if self.message == None:
            return None
        
        return self.message.content
    

class ChatTitleEditPayload:
    def __init__(self,
        timestamp: int,
        user: User,
        chat_id: "int | None" = None,
        title: "str | None" = None
    ):
        '''
        Payload that is sent to the `Bot.on_chat_title_change` decorator.

        :param timestamp: Timestamp of the title edit.
        :param user: User that edited the chat name.
        :param chat_id: Chat ID that had its title edited.
        :param title: New chat title
        '''
        self.timestamp: int = timestamp
        self.user: User = user
        self.chat_id: "int | None" = chat_id
        self.title: "str | None" = title
    
    
    @property
    def user_id(self):
        return self.user.user_id


    @staticmethod
    def from_json(data: dict) -> "ChatTitleEditPayload | None":
        if data == None: return None
        
        return ChatTitleEditPayload(
            data['timestamp'],
            User.from_json(data['user']),
            data.get('chat_id', None),
            data.get('title', None),
        )


class ChatMembershipPayload:
    def __init__(self,
        timestamp: int,
        user: User,
        chat_id: "int | None" = None,
        is_channel: bool = False
    ):
        '''
        Payload that is sent to the `Bot.on_bot_add` or `Bot.on_bot_remove` decorator.

        :param timestamp: Timestamp of the action.
        :param user: User that invited or kicked the bot.
        :param chat_id: Chat ID that the bot was invited to / kicked from.
        :param is_channel: Whether the bot got added to / kicked from a channel or not
        '''
        self.timestamp: int = timestamp
        self.user: User = user
        self.chat_id: "int | None" = chat_id
        self.is_channel: bool = is_channel
    

    @property
    def user_id(self):
        return self.user.user_id


    @staticmethod
    def from_json(data: dict) -> "ChatMembershipPayload | None":
        if data == None: return None
        
        return ChatMembershipPayload(
            data['timestamp'],
            User.from_json(data['user']),
            data.get('chat_id', None),
            data.get('is_channel', False),
        )


class UserMembershipPayload:
    def __init__(self,
        timestamp: int,
        user: User,
        chat_id: "int | None" = None,
        is_channel: bool = False,
        initiator: "int | None" = None
    ):
        '''
        Payload that is sent to the `Bot.on_user_add` or `Bot.on_user_remove` decorator.

        :param timestamp: Timestamp of the action.
        :param user: User that joined or left the chat.
        :param chat_id: Chat ID that the user joined / left.
        :param is_channel: Whether the user was added to / kicked from a channel or not.
        :param initiator: User ID of the inviter / kicker, if the user got invited by another user or kicked by an admin.
        '''
        self.timestamp: int = timestamp
        self.user: User = user
        self.chat_id: "int | None" = chat_id
        self.is_channel: bool = is_channel
        self.initiator: "int | None" = initiator


    @property
    def user_id(self):
        return self.sender.user_id


    @staticmethod
    def from_json(data: dict) -> "UserMembershipPayload | None":
        if data == None: return None
        
        return UserMembershipPayload(
            data['timestamp'],
            User.from_json(data['user']),
            data.get('chat_id', None),
            data.get('is_channel', False),
            data.get('inviter_id', data.get('admin_id', None))
        )



from . import types
import re
from typing import *

class _OrFilter:
    '''
    Class for using bit-wise or on filters
    '''
    def __init__(self, filter1, filter2):
        self.filter1 = filter1
        self.filter2 = filter2 if not isinstance(filter2, str) else equals(filter2)

    def __call__(self, obj: any):
        return self.filter1(obj) or self.filter2(obj)
    

class _AndFilter:
    '''
    Class for using bit-wise and on filters
    '''
    def __init__(self, filter1, filter2):
        self.filter1 = filter1
        self.filter2 = filter2 if not isinstance(filter2, str) else equals(filter2)

    def __call__(self, obj: any):
        return self.filter1(obj) and self.filter2(obj)


class _filter:
    '''
    Superclass of other filters for support of bit-wise or and bit-wise and
    '''
    def __or__(self, other):
        return _OrFilter(self, other)
    
    def __and__(self, other):
        return _AndFilter(self, other)


class equals(_filter):
    def __init__(self, content: str):
        '''
        :param content: Content to check
        
        Checks if the content equals to the given string
        '''
        self.content = content
    
    def __call__(self, obj: any):
        if hasattr(obj, "content"):
            return obj.content == self.content
        else:
            raise Exception(f"Class {type(object).__name__} has no content")


class has(_filter):
    def __init__(self, content: str):
        '''
        :param content: Content to check
        
        Checks if the content has the given string
        '''
        self.content = content
    
    def __call__(self, obj: any):
        if hasattr(obj, "content"):
            return self.content in obj.content
        else:
            raise Exception(f"Class {type(object).__name__} has no content")


class startswith(_filter):
    def __init__(self, prefix: str):
        '''
        :param prefix: Prefix to check
        
        Checks if the content starts with the given prefix
        '''
        self.prefix = prefix
    
    def __call__(self, obj: any):
        if hasattr(obj, "content"):
            return obj.content.startswith(self.prefix)
        else:
            raise Exception(f"Class {type(object).__name__} has no content")
        

class endswith(_filter):
    def __init__(self, suffix: str):
        '''
        :param suffix: Suffix to check
        
        Checks if the content ends with the given suffix
        '''
        self.suffix = suffix
    
    def __call__(self, obj: any):
        if hasattr(obj, "content"):
            return obj.content.endswith(self.suffix)
        else:
            raise Exception(f"Class {type(object).__name__} has no content")
        

class regex(_filter):
    def __init__(self, pattern: str):
        '''
        :param pattern: Regex pattern to check
        
        Checks if the content matches the given pattern
        '''
        self.pattern = pattern
    
    def __call__(self, obj: any):
        if hasattr(obj, "content"):
            return re.fullmatch(self.pattern, obj.body.text)
        else:
            raise Exception(f"Class {type(obj).__name__} has no content")


def papaya(obj: any):
    '''
    Checks if the content's second-to-last word of the content is "папайя".

    You do not need to call this.
    '''
    if hasattr(obj, "content"):
        words = obj.content.split()
        if len(words) < 2: return False
        return words[-2].lower() == "папайя"
    else:
        raise Exception(f"Class {type(object).__name__} has no content")

class state(_filter):
    def __init__(self, state: Any):
        '''
        :param state: State to check
        
        Checks if the content matches the given pattern
        '''
        self.state = state
    
    def __call__(self, obj: Any):
        if not hasattr(obj, "user_id"):
            raise Exception(f"Class {type(object).__name__} has no user id")
        
        user_id = obj.user_id

        if not user_id:
            return False
        
        if not hasattr(obj, "bot") or not obj.bot:
            return False
        
        storage = obj.bot.storage

        return storage.get_state(user_id) == self.state


import asyncio
from typing import *
import aiohttp

from . import utils
from .types import *
from .cache import *
from . import fsm
from .router import *
from . import exceptions

import logging
bot_logger = logging.getLogger("aiomax.bot")

class Bot(Router):
    def __init__(self,
        access_token: str,
        command_prefixes: "str | List[str]" = "/",
        mention_prefix: bool = True,
        case_sensitive: bool = True,
        default_format: "Literal['markdown', 'html'] | None" = None,
        max_messages_cached: int = 10000
    ):
        '''
        Bot init

        :param access_token: Bot access token from https://max.ru/masterbot
        :param command_prefixes: List of command prefixes or a command prefix
        :param mention_prefix: Whether to respond to commands starting with the ping of the bot
        :param case_sensitive: If False the bot will respond to commands regardless of case
        :param default_format: Default message formatting mode
        :param max_messages_cached: Maximum number of messages to cache. Set to 0 to disable caching
        '''
        super().__init__(case_sensitive)

        self.access_token: str = access_token
        self.session = None
        self.polling = False

        self.command_prefixes: "str | List[str]" = command_prefixes
        self.mention_prefix: bool = mention_prefix
        self.default_format: "str | None" = default_format
        self.cache: "MessageCache | None" = MessageCache(max_messages_cached)\
            if max_messages_cached > 0 else None
        
        self.id: "int | None" = None
        self.username: "str | None" = None
        self.name: "str | None" = None
        self.description: "str | None" = None
        self.bot_commands: List[BotCommand] = None

        self.marker: "int | None" = None

        self.storage = fsm.FSMStorage()

    async def get(self, *args, **kwargs):
        '''
        Sends a GET request to the API.
        '''
        if self.session == None:
            raise Exception("Session is not initialized")
        
        params = kwargs.get('params', {})
        params['access_token'] = self.access_token
        if 'params' in kwargs:
            del kwargs['params']
        
        response = await self.session.get(*args, params=params, **kwargs)

        if response.status == 200:
            return response
    
        raise await utils.get_exception(response)
    
    
    async def post(self, *args, **kwargs):
        '''
        Sends a POST request to the API.
        '''
        if self.session == None:
            raise Exception("Session is not initialized")
        
        params = kwargs.get('params', {})
        params['access_token'] = self.access_token
        if 'params' in kwargs:
            del kwargs['params']
       
        response = await self.session.post(*args, params=params, **kwargs)

        if response.status == 200:
            return response
    
        raise await utils.get_exception(response)
    
    
    async def patch(self, *args, **kwargs):
        '''
        Sends a PATCH request to the API.
        '''
        if self.session == None:
            raise Exception("Session is not initialized")
        
        params = kwargs.get('params', {})
        params['access_token'] = self.access_token
        if 'params' in kwargs:
            del kwargs['params']
        
        response = await self.session.patch(*args, params=params, **kwargs)

        if response.status == 200:
            return response
    
        raise await utils.get_exception(response)
    
    
    async def put(self, *args, **kwargs):
        '''
        Sends a PUT request to the API.
        '''
        if self.session == None:
            raise Exception("Session is not initialized")
        
        params = kwargs.get('params', {})
        params['access_token'] = self.access_token
        if 'params' in kwargs:
            del kwargs['params']
        
        response = await self.session.put(*args, params=params, **kwargs)

        if response.status == 200:
            return response
    
        raise await utils.get_exception(response)
    
    
    async def delete(self, *args, **kwargs):
        '''
        Sends a DELETE request to the API.
        '''
        if self.session == None:
            raise Exception("Session is not initialized")
        
        params = kwargs.get('params', {})
        params['access_token'] = self.access_token
        if 'params' in kwargs:
            del kwargs['params']
        
        response = await self.session.delete(*args, params=params, **kwargs)

        if response.status == 200:
            return response
    
        raise await utils.get_exception(response)
    

    # send requests

    async def get_me(self) -> User:
        '''
        Returns info about the bot.
        '''
        response = await self.get(f"https://botapi.max.ru/me")
        user = await response.json()
        user = User.from_json(user)

        # caching info
        self.id = user.user_id
        self.username = user.username
        self.name = user.name
        self.bot_commands = user.commands
        self.description = user.description
        return user


    async def patch_me(
        self,
        name: "str | None" = None,
        description: "str | None" = None,
        commands: "List[BotCommand] | None" = None,
        photo: "ImageRequestPayload | None" = None
    ) -> User:
        '''
        Allows you to change info about the bot. Fill in only the fields that
        need to be updated.
        
        :param name: Bot display name
        :param description: Bot description
        :param commands: Commands supported by the bot. To remove all commands,
        pass an empty list.
        :param photo: Bot profile pictur
        '''
        if commands:
            commands = [i.as_dict() for i in commands]   
        if photo:
            photo = photo.as_dict()
        
        payload = {
            "name": name,
            "description": description,
            "commands": commands,
            "photo": photo
        }
        payload = {k: v for k, v in payload.items() if v}

        response = await self.patch(f"https://botapi.max.ru/me", json=payload)
        data = await response.json()
    
        # caching info
        if name:
            self.name = name
        if commands:
            self.bot_commands = commands
        if description:
            self.description = description
        
        if response.status == 200:
            return User.from_json(data)
        else:
            bot_logger.error(f"Failed to update bot info: {data}. ")
    
    
    async def get_chats(self, count: "int | None" = None, marker: "int | None" = None) -> List[Chat]:
        '''
        Returns the chats the bot is in.
        The result includes a list of chats and a marker for moving to the next page.

        :param count:  Number of chats requested. 50 by default
        :param marker: Pointer to the next page of data. Defaults to first page
        '''
        params = {
            "count": count,
            "marker": marker,
        }
        params = {k: v for k, v in params.items() if v}

        response = await self.get("https://botapi.max.ru/chats", params=params)
        data = await response.json()
        chats = [Chat.from_json(i) for i in data['chats']]

        return chats
    

    async def chat_by_link(self, link: str) -> Chat:
        '''
        Returns chat by a link or username.
 
        :param link: Public chat link or username.
        '''
        response = await self.get(f"https://botapi.max.ru/chats/{link}")
        json = await response.json()

        return Chat.from_json(json)
    
    
    async def get_chat(self, chat_id: int) -> Chat:
        '''
        Returns information about a chat.

        :param chat_id: The ID of the chat.
        '''
        response = await self.get(f"https://botapi.max.ru/chats/{chat_id}")
        json = await response.json()

        return Chat.from_json(json)
    
    
    async def get_pin(self, chat_id: int) -> "Message | None":
        '''
        Returns pinned message in the chat as ``. None if there is no pinned message

        :param chat_id: The ID of the chat.
        '''
        response = await self.get(f"https://botapi.max.ru/chats/{chat_id}/pin")
        json = await response.json()

        if json['message'] == None:
            return None

        return Message.from_json(json)


    async def pin(self,
        chat_id: int,
        message_id: str,
        notify: "bool | None" = None
    ):
        '''
        Pin a message in a chat

        :param chat_id: The ID of the chat.
        :param message_id: The ID of the message to pin.
        :param notify: Whether to notify users about the pin. True by default.
        '''
        payload = {
            "message_id": message_id,
            "notify": notify
        }
        payload = {k: v for k, v in payload.items() if v}

        response = await self.put(
            f"https://botapi.max.ru/chats/{chat_id}/pin", json=payload
        )
        return await response.json()


    async def delete_pin(self, chat_id: int):
        '''
        Delete pinned message in the chat

        :param chat_id: The ID of the chat.
        '''
        response = await self.delete(f"https://botapi.max.ru/chats/{chat_id}/pin")

        return await response.json()
    
    
    async def my_membership(self, chat_id: int) -> User:
        '''
        Returns information about the bot's membership in the chat.

        :param chat_id: The ID of the chat.
        '''
        response = await self.get(f"https://botapi.max.ru/chats/{chat_id}/members/me")
        json = await response.json()

        return User.from_json(json)


    async def leave_chat(self, chat_id: int):
        '''
        Remove the bot from the chat.

        :param chat_id: The ID of the chat.
        '''
        response = await self.delete(f"https://botapi.max.ru/chats/{chat_id}/members/me")

        return await response.json()


    async def get_admins(self, chat_id: int) -> List[User]:
        '''
        Returns a list of administrators in the chat.

        :param chat_id: The ID of the chat.
        '''
        response = await self.get(f"https://botapi.max.ru/chats/{chat_id}/members/admins")

        users = [User.from_json(i) for i in (await response.json())['members']]

        return users
    

    async def get_memberships(self, chat_id: int, user_ids: "List[int] | int") -> "List[User] | User | None":
        '''
        Returns a list of memberships in the chat for the users with the specified ID.
        '''
        params = {
            'user_ids': user_ids if type(user_ids) == list else [user_ids]
        }
        response = await self.get(f"https://botapi.max.ru/chats/{chat_id}/members", params=params)

        users = [User.from_json(i) for i in (await response.json())['members']]

        if type(user_ids) == list:
            return users
        else:
            return users[0] if len(users) > 0 else None


    async def get_members(self, chat_id: int, count_per_iter: int = 100) -> AsyncIterator[User]:
        '''
        Returns an asynchronous interator of members in the chat.

        :param chat_id: The ID of the chat.
        :param count_per_iter: The number of users to fetch per request.
        '''
        marker = None
        
        while True:
            params = {
                "count": count_per_iter,
                "marker": marker,
            }
            params = {k: v for k, v in params.items() if v}
            response = await self.get(f"https://botapi.max.ru/chats/{chat_id}/members", params=params)
            data = await response.json()

            users = [User.from_json(i) for i in data['members']]

            for user in users:
                yield user

            marker = data.get('marker', None)
            if marker == None:
                break

    
    async def add_members(self,
        chat_id: int,
        users: List[int]
    ):
        '''
        Adds users to the chat.

        :param chat_id: The ID of the chat.
        :param users: List of user IDs to add.
        '''

        response = await self.post(f"https://botapi.max.ru/chats/{chat_id}/members", json={"user_ids": users})

        return await response.json()
  

    async def kick_member(
        self,
        chat_id: int,
        user_id: int,
        block: "bool | None" = None
    ):
        '''
        Removes a user from the chat.

        :param chat_id: The ID of the chat.
        :param user_id: The ID of the user to remove.
        :param block: Whether to block the user. Ignored by default.
        '''

        params = {
            "chat_id": chat_id,
            "user_id": user_id,
            "block": block
        }
        params = {k: v for k, v in params.items() if v}
        
        if block != None:
            params["block"] = str(block)

        response = await self.delete(f"https://botapi.max.ru/chats/{chat_id}/members/", params=params)

        return await response.json()
    

    async def patch_chat(self,
        chat_id: int,
        icon: "ImageRequestPayload | None" = None,
        title: "str | None" = None,
        pin: "str | None" = None,
        notify: "bool | None" = None
    ) -> Chat:
        '''
        Allows you to edit chat information, like the name,
        icon and pinned message.

        :param chat_id: ID of the chat to change
        :param icon: Chat picture
        :param title: Chat name. From 1 to 200 characters
        :param pin: ID of the message to pin
        :param notify: Whether to notify users about the edit. True by default.
        '''

        payload = {
            "icon": icon.as_dict() if icon else None,
            "title": title,
            "pin": pin,
            "notify": notify
        }
        payload = {k: v for k, v in payload.items() if v}

        response = await self.patch(
            f"https://botapi.max.ru/chats/{chat_id}", json=payload
        )
        json = await response.json()

        return Chat.from_json(json)
    

    async def post_action(self, chat_id: int, action: str):
        '''
        Allows you to show a badge about performing an action in a chat, like
        "typing". Also allows for marking messages as read.
        
        :param chat_id: ID of the chat to do the action in
        :param action: The action to perform
        '''

        response = await self.post(f"https://botapi.max.ru/chats/{chat_id}/actions", json={"action": action})

        return await response.json()
    

    async def upload(self, data: "IO | str", type: str) -> dict:
        '''
        Uploads a file to the server. Returns raw JSON with the token.

        You probably should use one of the upload_* methods instead.

        :param data: File-like object or path to the file
        :param type: File type
        '''
        if isinstance(data, str):
            data = open(data, 'rb')
        
        form = aiohttp.FormData()
        form.add_field('data', data)
        url_resp = await self.post('https://botapi.max.ru/uploads', params={"type": type})
        url_json = await url_resp.json()
        token_resp = await self.session.post(url_json['url'], data=form)
        if type in {'audio', 'video'}:
            return url_json
        token_json = await token_resp.json()

        return token_json
    

    async def upload_image(self, data: "BinaryIO | str") -> PhotoAttachment:
        '''
        Uploads an image to the server and returns a PhotoAttachment.

        :param data: File-like object or path to the file
        '''
        raw_photo = await self.upload(data, 'image')
        token = list(raw_photo['photos'].values())[0]['token']
        return PhotoAttachment(PhotoPayload(token=token))
    

    async def upload_video(self, data: "BinaryIO | str") -> VideoAttachment:
        '''
        Uploads a video to the server and returns a VideoAttachment.

        :param data: File-like object or path to the file
        '''
        raw_video = await self.upload(data, 'video')
        token = raw_video['token']
        return VideoAttachment(MediaPayload(token=token))
    

    async def upload_audio(self, data: "BinaryIO | str") -> AudioAttachment:
        '''
        Uploads an audio file to the server and returns an AudioAttachment.

        :param data: File-like object or path to the file
        '''
        raw_audio = await self.upload(data, 'audio')
        token = raw_audio['token']
        return AudioAttachment(MediaPayload(token=token))
    

    async def upload_file(self, data: "IO | str") -> FileAttachment: 
        '''
        Uploads a file to the server and returns a FileAttachment.

        :param data: File-like object or path to the file
        '''
        raw_file = await self.upload(data, 'file')
        token = raw_file['token']
        return FileAttachment(MediaPayload(token=token))


    async def send_message(self,
        text: str,
        chat_id: "int | None" = None,
        user_id: "int | None" = None,
        format: "Literal['markdown', 'html', 'default'] | None" = 'default',
        reply_to: "int | None" = None,
        notify: bool = True,
        disable_link_preview: bool = False,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> Message:
        '''
        Allows you to send a message to a user or in a chat.
        
        :param text: Message text. Up to 4000 characters
        :param chat_id: Chat ID to send the message in.
        :param user_id: User ID to send the message to.
        :param format: Message format. Bot.default_format by default
        :param reply_to: ID of the message to reply to. Optional
        :param notify: Whether to notify users about the message. True by default.
        :param disable_link_preview: Whether to disable link embedding in messages. True by default
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        # error checking
        text = str(text)
        assert chat_id or user_id, "Either chat_id or user_id must be provided"
        assert not (chat_id and user_id), "Both chat_id and user_id cannot be provided"

        # sending
        params = {
            "chat_id": chat_id,
            "user_id": user_id,
            "disable_link_preview": str(disable_link_preview).lower()
        }
        params = {k: v for k, v in params.items() if v}

        if format == 'default':
            format = self.default_format

        body = utils.get_message_body(text, format, reply_to, notify, keyboard, attachments)

        try:
            response = await self.post(
                f"https://botapi.max.ru/messages", params=params, json=body
            )
            json = await response.json()
            message = Message.from_json(json['message'])
            message.bot = self
            return message
        
        except exceptions.AttachmentNotReady:
            await asyncio.sleep(1)
            return await self.send_message(text=text, chat_id=chat_id, user_id=user_id, format=format, reply_to=reply_to, notify=notify, disable_link_preview=disable_link_preview, attachments=attachments)


    async def edit_message(self,
        message_id: str,
        text: "str | None" = None,
        format: "Literal['markdown', 'html', 'default'] | None" = 'default',
        reply_to: "int | None" = None,
        notify: bool = True,
        keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
        attachments: "List[Attachment] | Attachment | None" = None
    ) -> Message:
        '''
        Allows you to edit a message.
        
        :param message_id: ID of the message to edit
        :param text: Message text. Up to 4000 characters
        :param format: Message format. Bot.default_format by default
        :param reply_to: ID of the message to reply to. Optional
        :param notify: Whether to notify users about the message. True by default.
        :param keyboard: An inline keyboard to attach to the message
        :param attachments: List of attachments
        '''
        # editing
        params = {
            "message_id": message_id
        }
        if format == 'default':
            format = self.default_format
            
        body = utils.get_message_body(text, format, reply_to, notify, keyboard, attachments)

        response = await self.put(
            f"https://botapi.max.ru/messages", params=params, json=body
        )
        
        json = await response.json()
        if not json['success']:
            raise Exception(json['message'])
        message = Message.from_json(json)
        message.bot = self
        return message


    async def delete_message(self,
        message_id: str
    ):
        '''
        Allows you to delete a message in chat.
        
        :param message_id: ID of the message to delete
        '''
        # editing
        params = {
            "message_id": message_id
        }

        response = await self.delete(
            f"https://botapi.max.ru/messages", params=params
        )
        
        json = await response.json()
        if not json['success']:
            raise Exception(json['message'])


    # async def get_message(self,
    #     message_id: str
    # ) -> Message:
    #     '''
    #     Allows you to fetch message's info.
        
    #     :param message_id: ID of the message to get info of
    #     '''
    #     assert message_id.startswith('mid.'), "Message ID invalid"

    #     message_id = message_id[4:]

    #     # editing
    #     response = await self.get(
    #         f"https://botapi.max.ru/messages/{message_id}"
    #     )
    #     if response.status != 200:
    #         raise Exception(await response.text())
        
    #     return Message.from_json(await response.json())
    # todo fix - for some reason API replies with "invalid message_id"


    async def get_updates(self, limit: int = 100) -> tuple[int, dict]:
        '''
        Get bot updates / events.
        
        :param limit: Maximum amount of updates to return.
        '''
        payload = {
            "limit": limit,
            "marker": self.marker
        }
        payload = {k: v for k, v in payload.items() if v}

        response = await self.get(
            f"https://botapi.max.ru/updates", params=payload
        )
        json = await response.json()
        if 'marker' in json:
            self.marker = json['marker']

        return json
    

    async def handle_update(self, update: dict):
        '''
        Handles an update.
        '''
        update_type = update['update_type'] 

        if update_type == "message_created":
            message = Message.from_json(update["message"])
            message.bot = self
            message.user_locale = update.get('user_locale', None)
            cursor = fsm.FSMCursor(self.storage, message.sender.user_id)

            # caching
            if self.cache:
                self.cache.add_message(message)

            # handling commands
            prefixes = self.command_prefixes if type(self.command_prefixes) != str\
                else [self.command_prefixes]
            prefixes = list(prefixes)
            handled = False

            if self.mention_prefix:
                prefixes.extend([f'@{self.username} {i}' for i in prefixes])

            for prefix in prefixes:
                if len(message.body.text) <= len(prefix):
                    continue

                prefix = prefix if self.case_sensitive else prefix.lower()
                if not message.body.text.startswith(prefix):
                    continue

                command = message.body.text[len(prefix):]
                name = command.split()[0]
                check_name = name if self.case_sensitive else name.lower()
                args = ' '.join(command.split()[1:])
                
                if check_name not in self.commands:
                    bot_logger.debug(f"Command \"{name}\" not handled")
                    return

                if len(self.commands[check_name]) == 0:
                    bot_logger.debug(f"Command \"{name}\" not handled")
                    return

                block = False
                
                for i in self.commands[check_name]:
                    kwargs = utils.context_kwargs(i.call, cursor=cursor)
                    asyncio.create_task(i.call(CommandContext(
                        self, message, name, args
                    ), **kwargs))

                    if not i.as_message:
                        block = True

                bot_logger.debug(f"Command \"{name}\" handled")
                if block:
                    return

            # handling
            handled = False

            for handler in self.handlers['message_created']:
                filters = [filter(message) for filter in handler.filters]

                if all(filters):
                    kwargs = utils.context_kwargs(handler.call, cursor=cursor)
                    asyncio.create_task(handler.call(message, **kwargs))
                    handled = True
            
            # handle logs
            if handled:
                bot_logger.debug(f"Message \"{message.body.text}\" handled")
            else:
                bot_logger.debug(f"Message \"{message.body.text}\" not handled")


        if update_type == 'message_edited': 
            message = Message.from_json(update["message"])
            message.bot = self
            message.user_locale = update.get('user_locale', None)
            cursor = fsm.FSMCursor(self.storage, message.sender.user_id)

            # caching
            old_message = None
            if self.cache:
                old_message = self.cache.get_message(message.id)
                self.cache.add_message(message)

            # handling
            for handler in self.handlers[update_type]:
                filters = [filter(message) for filter in handler.filters]

                if all(filters):
                    kwargs = utils.context_kwargs(handler.call, before=old_message, after=message, cursor=cursor)
                    asyncio.create_task(handler.call(old_message, message, **kwargs))

            # handle logs
            bot_logger.debug(f"Message \"{message.body.text}\" edited")


        if update_type == 'message_removed': 
            payload = MessageDeletePayload.from_json(update, self)

            if payload.user_id:
                cursor = fsm.FSMCursor(self.storage, payload.user_id)
            else:
                cursor = None

            # handling
            for handler in self.handlers[update_type]:
                filters = [filter(payload) for filter in handler.filters]

                if all(filters):
                    kwargs = utils.context_kwargs(handler.call, cursor=cursor)
                    asyncio.create_task(handler.call(payload, **kwargs))

            # handle logs
            bot_logger.debug(f"Message \"{payload.content}\" deleted")

                
        if update_type == 'bot_started':
            payload = BotStartPayload.from_json(update, self)
            cursor = fsm.FSMCursor(self.storage, payload.user.user_id)

            bot_logger.debug(f"User \"{payload.user!r}\" started bot")

            for i in self.handlers[update_type]:
                kwargs = utils.context_kwargs(i, cursor=cursor)
                asyncio.create_task(i(payload, **kwargs))

                
        if update_type == 'chat_title_changed':
            payload = ChatTitleEditPayload.from_json(update)
            cursor = fsm.FSMCursor(self.storage, payload.user.user_id)

            bot_logger.debug(f"User \"{payload.user!r}\" changed title of chat {payload.chat_id}")

            for i in self.handlers[update_type]:
                kwargs = utils.context_kwargs(i, cursor=cursor)
                asyncio.create_task(i(payload, **kwargs))

                
        if update_type == 'bot_added' or update_type == 'bot_removed':
            payload = ChatMembershipPayload.from_json(update)
            cursor = fsm.FSMCursor(self.storage, payload.user.user_id)

            for i in self.handlers[update_type]:
                kwargs = utils.context_kwargs(i, cursor=cursor)
                asyncio.create_task(i(payload, **kwargs))
                
                
        if update_type == 'user_added' or update_type == 'user_removed':
            payload = UserMembershipPayload.from_json(update)
            cursor = fsm.FSMCursor(self.storage, payload.user.user_id)

            for i in self.handlers[update_type]:
                kwargs = utils.context_kwargs(i, cursor=cursor)
                asyncio.create_task(i(payload, **kwargs))

                
        if update_type == 'message_callback':
            handled = False

            callback = Callback.from_json(
                update['callback'], update.get('message', None), update.get('user_locale', None), self
            )

            cursor = fsm.FSMCursor(self.storage, callback.user.user_id)
  
            for handler in self.handlers[update_type]:
                filters = [filter(callback) for filter in handler.filters]

                if all(filters):
                    kwargs = utils.context_kwargs(handler.call, cursor=cursor)
                    asyncio.create_task(handler.call(callback, **kwargs))
                    handled = True
                
            if handled:
                bot_logger.debug(f"Callback \"{callback.payload}\" handled")
            else:
                bot_logger.debug(f"Callback \"{callback.payload}\" not handled")

                
        if update_type == 'message_chat_created':
            payload = ChatCreatePayload.from_json(update)
            bot_logger.debug(f"Created chat \"{payload.start_payload}\"")

            for i in self.handlers[update_type]:
                asyncio.create_task(i(payload))


    async def start_polling(self, session: "aiohttp.ClientSession | None" = None):
        '''
        Starts polling.

        :param session: Custom aiohttp client session
        '''
        self.polling = True

        if not session:
            session = aiohttp.ClientSession()

        async with session:
            self.session = session

            # self info (this will cache the info automatically)
            await self.get_me()

            bot_logger.info(f"Started polling with bot @{self.username} ({self.id}) - {self.name}")

            # ready event
            for i in self.handlers['on_ready']:
                await i()

            while self.polling:
                try:
                    updates = await self.get_updates()
                    
                    for update in updates["updates"]:
                        await self.handle_update(update)

                except Exception as e:
                    bot_logger.exception(e)
                    await asyncio.sleep(3)

        self.session = None
        self.polling = False


    def run(self, *args, **kwargs):
        '''
        Shortcut for `asyncio.run(Bot.start_polling())`
        '''
        asyncio.run(self.start_polling(*args, **kwargs))


from typing import *
from . import types


class MessageCache:
    def __init__(self, max_size: int = 10000):
        '''
        New message cache.

        :param max_size: Maximum number of messages to store
        '''
        self.max_size: int = max_size
        self.messages: Dict[str, types.Message] = {}


    def get_message(self, id: str) -> "types.Message | None":
        '''
        Returns a message by ID. None if message wasnt cached

        :param id: Message ID
        '''
        return self.messages.get(id, None)
    

    def add_message(self, message: types.Message):
        '''
        Caches a message.

        :param message: Message
        '''
        self.messages[message.id] = message

        while len(self.messages) > self.max_size:
            (k := next(iter(self.messages)), self.messages.pop(k))


from typing import *


class Button:
    def __init__(self,
        type: Literal['callback', 'link', 'request_geo_location', 'request_contact', 'chat'],
        text: str
    ):
        '''
        Base button class
        '''
        self.type: Literal['callback', 'link', 'request_geo_location', 'request_contact', 'chat'] = type
        self.text: str = text


    @staticmethod
    def from_json(data: dict) -> "Button":
        if data['type'] == 'callback':
            return CallbackButton.from_json(data)
        elif data['type'] == 'link':
            return LinkButton.from_json(data)
        elif data['type'] == 'request_geo_location':
            return GeolocationButton.from_json(data)
        elif data['type'] == 'request_contact':
            return ContactButton.from_json(data)
        elif data['type'] == 'chat':
            return ChatButton.from_json(data)
        else:
            raise Exception(f"Unknown button type: {data['type']}")
        

    def to_json(self) -> dict:
        return {
            'type': self.type,
            'text': self.text
        }


class CallbackButton(Button):
    def __init__(self,
        text: str,
        payload: str,
        intent: Literal['default', 'positive', 'negative'] = 'default'
    ):
        '''
        Callback button
        '''
        super().__init__('callback', text)
        self.payload: str = payload
        self.intent: Literal['default', 'positive', 'negative'] = intent

    
    @staticmethod
    def from_json(data: dict) -> "CallbackButton":
        return CallbackButton(data['text'], data['payload'], data['intent'])
    

    def to_json(self) -> dict:
        return {
            'type': 'callback',
            'text': self.text,
            'payload': self.payload,
            'intent': self.intent
        }


class LinkButton(Button):
    def __init__(self,
        text: str,
        url: str,
    ):
        '''
        URL button on a message

        :param text: Button text
        :param url: Link that the button redirects to
        '''
        super().__init__('link', text)
        self.url: str = url

    
    @staticmethod
    def from_json(data: dict) -> "LinkButton":
        return LinkButton(data['text'], data['url'])
    

    def to_json(self) -> dict:
        return {
            'type': 'link',
            'text': self.text,
            'url': self.url
        }


class GeolocationButton(Button):
    def __init__(self,
        text: str,
        quick: bool = False,
    ):
        '''
        Request geolocation button on a message

        :param text: Button text
        :param quick: Whether to show a confirmational message to a user when pressing the button
        '''
        super().__init__('request_geo_location', text)
        self.quick: bool = quick

    
    @staticmethod
    def from_json(data: dict) -> "GeolocationButton":
        return GeolocationButton(data['text'], data['quick'])
    

    def to_json(self) -> dict:
        return {
            'type': 'request_geo_location',
            'text': self.text,
            'quick': self.quick
        }


class ContactButton(Button):
    def __init__(self,
        text: str
    ):
        '''
        Request contact button on a message

        :param text: Button text
        '''
        super().__init__('request_contact', text)

    
    @staticmethod
    def from_json(data: dict) -> "ContactButton":
        return ContactButton(data['text'])
    

    def to_json(self) -> dict:
        return {
            'type': 'request_contact',
            'text': self.text
        }


class ChatButton(Button):
    def __init__(self,
        text: str,
        title: str,
        description: "str | None" = None,
        payload: "str | None" = None,
        uuid: "int | None" = None
    ):
        '''
        Chat creation button on a message

        :param text: Button text
        :param title: Name of the new chat
        :param description: Description of the new chat
        :param payload: Payload that will be sent to the bot when the chat is created
        :param uuid: Chat UUID, assigned when new message is sent. Provide when editing message 
        '''
        super().__init__('chat', text)
        self.title: str = title
        self.description: "str | None" = description
        self.payload: "str | None" = payload
        self.uuid: "int | None" = uuid

    
    @staticmethod
    def from_json(data: dict) -> "ChatButton":
        return ChatButton(
            data['text'],
            data.get('chat_title', None),
            data.get('chat_description', None),
            data.get('start_payload', None),
            data.get('uuid', None)
        )
    
    
    def to_json(self) -> dict:
        return {
            'type': 'chat',
            'text': self.text,
            'chat_title': self.title,
            'chat_description': self.description,
            'start_payload': self.payload,
            'uuid': self.uuid
        }


# builder

class KeyboardBuilder:
    def __init__(self):
        '''
        Keyboard builder
        '''
        self.buttons: List[List[Button]] = []


    def to_list(self) -> List[List[dict]]:
        '''
        Returns a serialised interpretation of the keyboard to put in a message
        '''
        return [[button.to_json() for button in row] for row in self.buttons]
    

    def add(self, *buttons: Button):
        '''
        Add buttons to the last row of the keyboard
        '''
        if len(self.buttons) == 0:
            self.buttons.append([])
        
        self.buttons[-1].extend(buttons)

    
    def row(self, *buttons: Button):
        '''
        Add a row of buttons
        '''
        self.buttons.append(list(buttons))
    

    def table(self,
        in_row: int,
        *buttons: Button
    ):
        '''
        Adds multiple rows of buttons so that there are `in_row` buttons in each row

        :param in_row: How many buttons to put in each row
        '''
        counter = 0

        for button in buttons:
            if counter == 0:
                self.row()
                counter = in_row

            self.add(button)
            counter -= 1
        


from typing import *

from aiomax.types import Attachment
from . import buttons
from inspect import signature
import aiohttp
from . import exceptions

def get_message_body(
    text: str,
    format: "Literal['markdown', 'html'] | None" = None,
    reply_to: "int | None" = None,
    notify: bool = True,
    keyboard: "List[List[buttons.Button]] | buttons.KeyboardBuilder | None" = None,
    attachments: "List[Attachment] | Attachment | None" = None
) -> dict:
    '''
    Returns the body of the message as json.
    '''
    body = {
        "text": text,
        "format": format,
        "notify": notify
    }

    # replying
    if reply_to:
        body['link'] = {
            "type": 'reply',
            "mid": reply_to
        }

    # keyboard
    if keyboard:
        if isinstance(keyboard, buttons.KeyboardBuilder):
            keyboard = keyboard.to_list()

        body['attachments'] = [{
            'type': 'inline_keyboard',
            'payload': {
                'buttons': [ [
                    i.to_json() if isinstance(i, buttons.Button) else i for i in row
                ] for row in keyboard ]
            }
        }]
    
    if attachments:
        if 'attachments' not in body:
            body['attachments'] = []

        if not isinstance(attachments, list):
            attachments = [attachments]

        for at in attachments or []:
            assert hasattr(at, 'as_dict'), 'Attachment must be an image, a video, an audio or a file'
            body['attachments'].append(at.as_dict())

    if attachments == [] and 'attachments' not in body:
        body['attachments'] = []

    return body

def context_kwargs(func: Callable, **kwargs):
    '''
    Returns only those kwargs, that callable accepts
    '''
    params = list(signature(func).parameters.keys())
    
    kwargs = {kw: arg for kw, arg in kwargs.items() if kw in params}

    return kwargs

async def get_exception(response: aiohttp.ClientResponse):
    if response.content_type == "text/plain":
        text = await response.text()
    elif response.content_type == "application/json":
        text = (await response.json())['code']
    else:
        return Exception(f"Unknown error: {await response.read()}")

    if text.startswith("Invalid access_token"):
        return exceptions.InvalidToken()
    if text == "attachment.not.ready":
        return exceptions.AttachmentNotReady()
    return Exception(f"Unknown error: {text}")



from setuptools import setup, find_packages

def readme():
    try:
        with open('pypi-description.md', encoding='utf-8') as f:
            return f.read()
    except:
        return 'Asynchronous framework for Max Bot API '

setup(
    name='aiomax',
    version='2.5.1',
    description='Asynchronous framework for Max Bot API ',
    author='oaa dpnspn',
    author_email='mbutsk@icloud.com',
    packages=find_packages(),
    install_requires=['aiohttp'],
    zip_safe=False,
    url = "https://github.com/dpnspn/aiomax",
    license="MIT License, see LICENSE.md file",
    long_description=readme(),
    long_description_content_type='text/markdown',
    keywords = [
        "bot",
        "api",
        "asyncio",
        "max"
    ],
    project_urls={
        "Docs": "https://github.com/dpnspn/aiomax/wiki"
    },
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Framework :: AsyncIO",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications :: Chat",
    ]
)




from copy import deepcopy
from typing import *
from .types import *
from .cache import *
from .filters import equals

import logging
bot_logger = logging.getLogger("aiomax.bot")


class Router:
    def __init__(self,
        case_sensitive: bool = True,
    ):
        '''
        Router init
t
        :param case_sensitive: If False the bot will respond to commands regardless of case
        '''
        self._handlers: dict[str, List[Handler]] = {
            'message_created': [],
            'on_ready': [],
            'bot_started': [],
            'message_callback': [],
            'message_chat_created': [],
            'message_edited': [],
            'message_removed': [],
            'chat_title_changed': [],
            'bot_added': [],
            'bot_removed': [],
            'user_added': [],
            'user_removed': []
        } # handlers in this router
        self._commands: dict[str, List[CommandHandler]] = {} # commands in this router
        self.case_sensitive: bool = case_sensitive
        self.parent = None # Parent bot of this router
        self.routers: List["Router"] = []
        self.filters: Dict[str, List[Callable]] = {
            'message_created': [],
            'message_edited': [],
            'message_removed': [],
            'message_callback': []
        }
        
    @staticmethod
    def wrap_filters(
        filters: Tuple["Callable | str | None", ...],
        mode: str = "and"
    ) -> Callable:
        '''
        Normalize multiple filters into a single callable.

        :param filters: filters to combine
        :param mode: "and" (default) — all filters must pass,
                     "or" — at least one filter must pass.
        '''
        normalized_filters = []

        for filter_ in filters:
            if filter_ is None:
                continue
            elif isinstance(filter_, str):
                normalized_filters.append(equals(filter_))
            elif callable(filter_):
                normalized_filters.append(filter_)
            else:
                raise ValueError(f"Unsupported filter type: {type(filter_)}")

        if not normalized_filters:
            return lambda message: True

        if mode == "and":
            def combined_filter(message):
                return all(f(message) for f in normalized_filters)
        elif mode == "or":
            def combined_filter(message):
                return any(f(message) for f in normalized_filters)
        else:
            raise ValueError(f"Unsupported mode: {mode}. Use 'and' or 'or'.")

        return combined_filter


    # routers

    @property
    def handlers(self) -> dict[str, List[Handler]]:
        '''
        Returns all handlers in this and all the child routers.
        '''
        out = deepcopy(self._handlers)

        for router in self.routers:
            for handler_type in out.keys():
                out[handler_type].extend(router.handlers[handler_type])

        return out
    

    @property
    def commands(self) -> dict[str, List[CommandHandler]]:
        '''
        Returns all commands in this and all the child routers.
        '''
        out = deepcopy(self._commands)
        for router in self.routers:
            out.update(router.commands)
        return out
    

    @property
    def bot(self) -> "Bot":
        '''
        Returns the bot this router is attached to.
        '''
        if self.parent == None:
            return None
        
        if self.parent.parent == None:
            return self.parent
        return self.parent.parent


    def add_router(self, router: "Router"):
        if router.parent != None:
            raise ValueError("Router already has a parent")
        
        router.parent = self
        self.routers.append(router)


    def remove_router(self, router: "Router"):
        if router not in self.routers:
            raise ValueError("Router not found")
        
        router.parent = None
        self.routers.remove(router)


    # decorators

    def on_message(self, *filters: "Callable | str | None", mode: str = 'and'):
        '''
        Decorator for receiving messages.
        '''
        def decorator(func):
            new_filter = self.wrap_filters(filters, mode=mode)

            self._handlers["message_created"].append(
                Handler(call=func, deco_filter=new_filter, router_filters=self.filters['message_created'])
            )
            return func
        return decorator


    def on_message_edit(self, *filters: "Callable | str | None", mode: str = 'and'):
        '''
        Decorator for editing messages.
        '''
        def decorator(func):
            new_filter = self.wrap_filters(filters, mode=mode)

            self._handlers["message_edited"].append(
                Handler(call=func, deco_filter=new_filter, router_filters=self.filters['message_edited'])
            )
            return func
        return decorator



    def on_message_delete(self, *filters: "Callable | str | None", mode: str = 'and'):
        '''
        Decorator for deleted messages.
        '''
        def decorator(func):
            new_filter = self.wrap_filters(filters, mode=mode)

            self._handlers["message_removed"].append(
                Handler(call=func, deco_filter=new_filter, router_filters=self.filters['message_removed'])
            )
            return func
        return decorator


    def on_bot_start(self):
        '''
        Decorator for handling bot start.
        '''
        def decorator(func): 
            self._handlers["bot_started"].append(func)
            return func
        return decorator


    def on_chat_title_change(self):
        '''
        Decorator for handling chat title changes.
        '''
        def decorator(func): 
            self._handlers["chat_title_changed"].append(func)
            return func
        return decorator


    def on_bot_add(self):
        '''
        Decorator for handling bot invitations in groups.
        '''
        def decorator(func): 
            self._handlers["bot_added"].append(func)
            return func
        return decorator


    def on_bot_remove(self):
        '''
        Decorator for handling bot kicks from groups.
        '''
        def decorator(func): 
            self._handlers["bot_removed"].append(func)
            return func
        return decorator


    def on_user_add(self):
        '''
        Decorator for handling user joins.
        '''
        def decorator(func): 
            self._handlers["user_added"].append(func)
            return func
        return decorator


    def on_user_remove(self):
        '''
        Decorator for handling user leaves.
        '''
        def decorator(func): 
            self._handlers["user_removed"].append(func)
            return func
        return decorator
    

    def on_ready(self):
        '''
        Decorator for receiving messages.
        '''
        def decorator(func): 
            self._handlers["on_ready"].append(func)
            return func
        return decorator


    def on_button_callback(self, *filters: "Callable | str | None", mode: str = 'and'):
        '''
        Decorator for receiving button presses.
        '''
        def decorator(func):
            new_filter = self.wrap_filters(filters, mode=mode)

            self._handlers["message_callback"].append(
                Handler(call=func, deco_filter=new_filter, router_filters=self.filters['message_callback'])
            )
            return func
        return decorator


    def on_button_chat_create(self):
        '''
        Decorator for receiving button presses.
        '''
        def decorator(func): 
            self._handlers["message_chat_created"].append(func)
            return func
        return decorator
    

    def on_command(self, name: "str | None" = None, aliases: List[str] = [], as_message: bool = False):
        '''
        Decorator for receiving commands.

        :param name: Command name
        :param aliases: List of alternative names for this command
        :param as_message: Whether to trigger on_message decorator when this command is invoked
        '''
        def decorator(func): 
            # command name
            if name is None:
                command_name = func.__name__
            else:
                assert ' ' not in name, 'Command name cannot contain spaces'
                command_name = name
            
            check_name = command_name.lower() if not self.case_sensitive else command_name
            if check_name not in self._commands:
                self._commands[check_name] = []
            self._commands[check_name].append(CommandHandler(func, as_message))

            # aliases
            for i in aliases:
                assert ' ' not in i, 'Command alias cannot contain spaces'

                check_name = i.lower() if not self.case_sensitive else i
                if check_name not in self._commands:
                    self._commands[check_name] = []
                self._commands[check_name].append(CommandHandler(func, as_message))
            return func
        return decorator
    

    # filters

    def add_message_filter(self, filter: "Callable"):
        self.filters['message_created'].append(filter)

    def add_message_edit_filter(self, filter: "Callable"):
        self.filters['message_edited'].append(filter)

    def add_message_delete_filter(self, filter: "Callable"):
        self.filters['message_removed'].append(filter)

    def add_button_callback_filter(self, filter: "Callable"):
        self.filters['message_callback'].append(filter)
